(*
    Copyright (C) 2012 Christophe Calvès

    This file is part of WC.

    WC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    WC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*)

open Terms
open Comp


(************************
 *                      *
 *       MESSAGES       *
 *                      *
 ************************)

let version =
"WC 0.0.1\n
Licensed under the terms of the GPLv3+\n\
Copyright GTof\n\
\n"


let usagemsg    =
"\
Usage: " ^ (Filename.basename Sys.argv.(0)) ^ " OPTIONS

 -t, --TeX                : compiles to TeX
 -s, --shell              : compiles to POSIX shell

 -a, --ast                : print ast
 -c, --core               : compiles to core
 -l, --alpha              : compiles to alpha renamed core
 -r, --reduced            : compiles to reduced core
 -S, --super              : compiles to supercombinators
 -p, --pseudo             : compiles to pseudo code

 -i, --intput <file path> : input  file ( or - for stdio )
 -o, --output <file path> : output file ( or - for stdout)
 -v, --version            : prints version informations

"

let texdesc     = "compiles to TeX"
let shelldesc   = "compiles to POSIX shell"

let astdesc     = "print abstract syntax tree"
let coredesc    = "compiles to core"
let alphdesc    = "compiled to alpha reduced core"
let reddesc     = "compiles to reduced core"
let superdesc   = "compiles to supercombinators"
let pseudodesc  = "compiled to pseudo code"

let versiondesc = "prints version information"
let inputdesc   = "input file path (or - for standard input)"
let outputdesc  = "output file path (or - for standard output)"



(************************
 *                      *
 *        PARAMS        *
 *                      *
 ************************)


let null _ = ()


(** Tells which method to run *)
type process = TeX
             | Shell
             | PAst
             | PCore
             | PAlpha
             | PReduce
             | PSuper
             | PPseudo
             | Version
             | Usage

(** reference that holds the method selected:
    by default: print usage
 *)
let process = ref Usage
let setProcess x _ = (process := x)

(** reference that holds the input, output file
    and recursion support.
    by default: startard input and output
                perform recursion
 *)
let inputfile  = ref "-"
let outputfile = ref "-"

(************************
 *                      *
 *       PROCESSING     *
 *                      *
 ************************)

(** lift a function f that takes an ast
    to a function that parses the input file
 *)
let upIn f () = 
    let (c,close) = (match !inputfile with
                       | "-" -> (stdin , null)
                       | x   -> let c = open_in x
                                in (c , (fun () -> close_in c))
                    )
 in let ast = Parser.ast Lexer.lexer (Lexing.from_channel c)
 in let _   = close ()
 in f ast


(** lift a function that returns a string to
    a functions that outputs to the output file
 *)
let upOut f x =
    let (c,close) = (match !outputfile with
                       | "-" -> (stdout , null)
                       | x   -> let c = open_out x
                                in (c , (fun () -> close_out c))
                    )
 in (output_string c (f x) ; flush c ; close ())


(************************
 *                      *
 *    COMPILATION       *
 *                      *
 ************************)

(** Print the ast
 *)
let processAst () =
 let f x = Pretty.pretty_run (pretty_ast x)
 in upIn (upOut f) ()


(** Compliled to core
 *)
let processCore () =
 let f x = Pretty.pretty_run (pretty_core (crun (coreOfAst x)))
 in upIn (upOut f) ()

(** Compliled to alpha renamed core
 *)
let processAlpha () =
 let f x = let m = coreOfAst x >>= coreAlpha
           in Pretty.pretty_run (pretty_core (crun m))
 in upIn (upOut f) ()


(** Compliled to reduced core
 *)
let processReduce () =
 let f x = let m = coreOfAst x >>= coreAlpha >>= (fun u ->
                   let _ = checkClos u in
                   simplifyRec u )
           in Pretty.pretty_run (pretty_core (crun m))
 in upIn (upOut f) ()


(** compiles an ast to a core *)
let core_of_ast x =
  coreOfAst x >>= coreAlpha >>= (fun u ->
                   let _ = checkClos u in
                   simplifyRec u >>= compileCore )

(** Compliled to super combinators
 *)
let processSuper () =
 let f x = let m = core_of_ast x >>= (fun main ->
                   superCombinators >>= (fun prog ->
                   creturn (Prog (prog , main)) ))
           in Pretty.pretty_run (pretty_prog (pretty_super string_of_int) (crun m))
 in upIn (upOut f) ()


(** Compliled to pseudo code
 *)
let processPseudo () =
 let f x = let m = core_of_ast x >>= (fun main ->
                   codes >>= (fun prog ->
                   creturn (Prog (prog , main)) ))
           in Pretty.pretty_run (pretty_prog pretty_codes (crun m))
 in upIn (upOut f) ()


(** return the path to the executable wc
    to know where are tex and shell directories
 *)
let exedir () =
    let s = Sys.getcwd ()
 in let _ = Sys.chdir (Filename.dirname Sys.argv.(0))
 in let r = Sys.getcwd ()
 in (Sys.chdir s ; r)

(** compiles the input file to TeX
 *)
let processTeX () =
 let f x = let m = core_of_ast x >>= (fun main ->
                   codes >>= (fun prog ->
                   creturn (Prog (prog , main)) ))
           and libpath = (exedir ()) ^ "/tex"
           in Tex.of_prog_standalone libpath (crun m)
 in upIn (upOut f) ()


(** compiles the input file to shell
 *)
let processShell () =
 let f x = let m = core_of_ast x >>= (fun main ->
                   codes >>= (fun prog ->
                   creturn (Prog (prog , main)) ))
           and libpath = (exedir ()) ^ "/shell"
           in Shell.of_prog_standalone libpath (crun m)
 in upIn (upOut f) ()



(************************
 *                      *
 *         MAIN         *
 *                      *
 ************************)

(** check the manual of module Arg for details *)
let main () =
 let _ =  Arg.parse [ ( "-t"        , Arg.Unit (setProcess TeX    ) , texdesc    )
                    ; ( "--TeX"     , Arg.Unit (setProcess TeX    ) , texdesc    )

                    ; ( "-s"        , Arg.Unit (setProcess Shell  ) , shelldesc  )
                    ; ( "--shell"   , Arg.Unit (setProcess Shell  ) , shelldesc  )

                    ; ( "-a"        , Arg.Unit (setProcess PAst  )  , astdesc    )
                    ; ( "--ast"     , Arg.Unit (setProcess PAst  )  , astdesc    )

                    ; ( "-c"        , Arg.Unit (setProcess PCore  ) , coredesc   )
                    ; ( "--core"    , Arg.Unit (setProcess PCore  ) , coredesc   )

                    ; ( "-l"        , Arg.Unit (setProcess PAlpha ) , coredesc   )
                    ; ( "--alpha"   , Arg.Unit (setProcess PAlpha ) , coredesc   )

                    ; ( "-r"        , Arg.Unit (setProcess PReduce) , coredesc   )
                    ; ( "--reduced" , Arg.Unit (setProcess PReduce) , coredesc   )

                    ; ( "-S"        , Arg.Unit (setProcess PSuper ) , superdesc  )
                    ; ( "--super"   , Arg.Unit (setProcess PSuper ) , superdesc  )

                    ; ( "-p"        , Arg.Unit (setProcess PPseudo) , pseudodesc )
                    ; ( "--pseudo"  , Arg.Unit (setProcess PPseudo) , pseudodesc )

                    ; ( "-i"        , Arg.Set_string inputfile      , inputdesc  )
                    ; ( "--input"   , Arg.Set_string inputfile      , inputdesc  )

                    ; ( "-o"        , Arg.Set_string outputfile     , outputdesc )
                    ; ( "--output"  , Arg.Set_string outputfile     , outputdesc )

                    ; ( "-v"        , Arg.Unit (setProcess Version) , versiondesc)
                    ; ( "--version" , Arg.Unit (setProcess Version) , versiondesc)
                    ]
                    null
                    usagemsg
 in let meth = (match !process with
                  | TeX     -> processTeX
                  | Shell   -> processShell
                  | PAst    -> processAst
                  | PCore   -> processCore
                  | PAlpha  -> processAlpha
                  | PReduce -> processReduce
                  | PSuper  -> processSuper
                  | PPseudo -> processPseudo
                  | Version -> upOut (Base.cst version )
                  | Usage   -> upOut (Base.cst usagemsg)
              )
 in meth ()
;;

main ()



